from ..base import EventIOFile, EventIOObject
from ..search_utils import yield_all_objects_depth_first
from argparse import ArgumentParser


parser = ArgumentParser()
parser.add_argument('inputfile')
parser.add_argument(
    '-m', '--max-repeats',
    default=5,
    type=int,
    help='If many of the same objects follow each other, only print the first `max-repeats` objects'
)


def main():
    args = parser.parse_args()

    with EventIOFile(args.inputfile) as f:
        last = None
        last_level = 0
        n_same = 0
        for o, level in yield_all_objects_depth_first(f):
            if last and last != EventIOObject and isinstance(o, last):
                n_same += 1
                last_level = level
            else:
                last = o.__class__
                if n_same > args.max_repeats:
                    print(
                        '    ' * last_level,
                        'And {} objects more of the same type'.format(
                            n_same - args.max_repeats
                        )
                    )
                n_same = 0
                last_level = level

            if n_same < args.max_repeats:
                print('    ' * level, o)


if __name__ == '__main__':
    main()



'''
Cleanly cut an eventio file, so that the
uncompressed size is at most `max_size`
'''
from argparse import ArgumentParser
import re
import gzip
from ..base import EventIOFile


parser = ArgumentParser(description=__doc__)
parser.add_argument('inputfile', help='Input eventio file')
parser.add_argument(
    'outputfile',
    help='Output file, if ending with .gz, it is written gzip compressed'
)
parser.add_argument(
    'max_size',
    help=(
        'Maximimum, uncompressed output file size.'
        ' You can use k, M, G suffixes.'
        ' As many complete top-level objects are written, so that max_size is'
        ' not exceeded'
    )
)


def parse_size(size):
    m = re.match(r'(\d+)([kMG])?', size)
    if not m:
        raise ValueError('Format not recognized, valid are #, #k, #M or #G')

    size, unit = m.groups()
    size = float(size)

    if unit is None:
        return int(size)

    if unit == 'k':
        return int(size * 1024)

    if unit == 'M':
        return int(size * 1024**2)

    if unit == 'G':
        return int(size * 1024**3)


def main():
    args = parser.parse_args()

    max_size = parse_size(args.max_size)
    bytes_to_read = 0
    with EventIOFile(args.inputfile) as f:
        last_o = None
        for o in f:
            last_byte = o.header.address + o.header.length
            if last_byte < max_size:
                bytes_to_read = last_byte
                last_o = o
            else:
                break

    print('Writing {:.2f}MB (uncompressed)'.format(bytes_to_read / 1024**2))
    print('Last object is of type {}'.format(last_o.header.type))

    if args.outputfile.endswith('.gz'):
        open_file = gzip.open
    else:
        open_file = open

    with open_file(args.outputfile, 'wb') as of:

        with EventIOFile(args.inputfile) as f:
            of.write(f.read(bytes_to_read))


if __name__ == '__main__':
    main()



from .base import EventIOFile, KNOWN_OBJECTS
from . import iact
from .iact import IACTFile
from . import simtel
from .simtel import SimTelFile
from .histograms import Histograms

__all__ = [
    'EventIOFile',
    'IACTFile',
    'SimTelFile',
    'simtel',
    'iact',
]


KNOWN_OBJECTS[Histograms.eventio_type] = Histograms



import struct
import numpy as np
from .var_int import get_length_of_varint, parse_varint


def read_short(f):
    ''' Read a signed 2 byte integer from `f`'''
    return struct.unpack('<h', f.read(2))[0]


def read_unsigned_short(f):
    ''' Read an unsigned 2 byte integer from `f`'''
    return struct.unpack('<H', f.read(2))[0]


def read_int(f):
    ''' Read a signed 4 byte integer from `f`'''
    return struct.unpack('<i', f.read(4))[0]


def read_unsigned_int(f):
    ''' Read an signed 4 byte integer from `f`'''
    return struct.unpack('<I', f.read(4))[0]


def read_float(f):
    ''' Read a 4 byte float from `f`'''
    return struct.unpack('<f', f.read(4))[0]


def read_array(f, dtype, count):
    '''Read a numpy array with `dtype` of length `count` from file-like `f`'''
    dt = np.dtype(dtype)
    if count == 0:
        return np.array((), dtype=dtype)
    return np.frombuffer(f.read(count * dt.itemsize), count=count, dtype=dt)


def read_eventio_string(f):
    '''Read a string from eventio file or object f.
    Eventio stores strings as a short giving the length
    of the string and the string itself.
    '''
    length = read_short(f)
    return f.read(length)


def read_from(f, fmt):
    '''
    read the struct fmt specification from file f
    Moves the current position.
    '''
    result = struct.unpack_from(
        fmt,
        f.read(struct.calcsize(fmt))
    )
    return result


def read_ints(f, n_ints):
    ''' read n ints from file f '''
    return read_from(f, '{:d}i'.format(n_ints))


def read_from_without_position_change(f, fmt):
    ''' Read struct format and return to old cursor position '''
    position = f.tell()
    result = read_from(f, fmt)
    f.seek(position)
    return result


def read_time(f):
    '''Read a time as combination of seconds and nanoseconds'''
    sec, nano = read_from(f, '<ii')
    return sec, nano


def read_utf8_like_signed_int(f):
    # this is mostly a verbatim copy from eventio.c lines 1082ff
    u = read_utf8_like_unsigned_int(f)
    # u values of 0,1,2,3,4,... here correspond to signed values of
    #   0,-1,1,-2,2,... We have to test the least significant bit:
    if (u & 1) == 1:  # Negative number;
        return -(u >> 1) - 1
    else:
        return u >> 1


def read_utf8_like_unsigned_int(f):
    '''this returns a python integer'''
    # this is a reimplementation from eventio.c lines 797ff
    var_int_bytes = bytearray(f.read(1))
    var_int_length = get_length_of_varint(var_int_bytes[0])
    if var_int_length - 1 > 0:
        var_int_bytes.extend(f.read(var_int_length - 1))

    return parse_varint(var_int_bytes)


def read_vector_of_uint32_scount_differential(f, count):
    return np.cumsum([read_utf8_like_signed_int(f) for _ in range(count)])


def read_vector_of_uint32_scount_differential_optimized(f, count):
    '''Stupid, pure python copy of eventio.c:1457'''
    output = np.empty(count, dtype='uint32')

    val = np.int32(0)
    for i in range(count):
        v0, = f.read(1)

        if (v0 & 0x80) == 0:  # one byte
            if (v0 & 1) == 0:  # positive
                val += v0 >> 1
            else:  # negative
                val -= (v0 >> 1) + 1
        elif (v0 & 0xc0) == 0x80:  # two bytes
            v1, = f.read(1)
            if (v1 & 1) == 0:  # positive
                val += ((v0 & 0x3f) << 7) | (v1 >> 1)
            else:  # negative
                val -= ((v0 & 0x3f) << 7) | ((v1 >> 1) + 1)
        elif (v0 & 0xe0) == 0xc0:  # three bytes
            v1, v2 = f.read(2)

            if (v2 & 1) == 0:
                val += (
                    ((v0 & 0x1f) << 15)
                    | (v1 << 7)
                    | (v2 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x1f) << 15)
                    | (v1 << 7)
                    | ((v2 >> 1) + 1)
                )
        elif (v0 & 0xf0) == 0xe0:  # four bytes
            v1, v2, v3 = f.read(3)
            if (v3 & 1) == 0:
                val += (
                    ((v0 & 0x0f) << 23)
                    | (v1 << 15)
                    | (v2 << 7)
                    | (v3 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x0f) << 23)
                    | (v1 << 15)
                    | (v2 << 7)
                    | ((v3 >> 1) + 1)
                )
        elif (v0 & 0xf8) == 0xf0:
            v1, v2, v3, v4 = f.read(4)
            # The format would allow bits 32 and 33 being set but we ignore this here. */
            if (v4 & 1) == 0:
                val += (
                    ((v0 & 0x07) << 31)
                    | (v1 << 23)
                    | (v2 << 15)
                    | (v3 << 7)
                    | (v4 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x07) << 31)
                    | (v1 << 23)
                    | (v2 << 15)
                    | (v3 << 7)
                    | ((v4 >> 1) + 1)
                )
        output[i] = val

    if count == 1:
        return val

    return output



from eventio import EventIOFile
from argparse import ArgumentParser
from collections import Counter
import json
import warnings
from eventio.simtel import TrackingPosition, TelescopeEvent


def count_versions(f):
    c = Counter()
    try:
        for o in f:
            if isinstance(o, TrackingPosition):
                eventio_type = 2100
            elif isinstance(o, TelescopeEvent):
                eventio_type = 2200
            else:
                eventio_type = o.header.type

            c.update([(
                eventio_type,
                o.header.version,
                o.__module__ + '.' + o.__class__.__qualname__)
            ])
            if o.header.only_subobjects:
                c.update(count_versions(o))
    except EOFError:
        warnings.warn("File seems to be truncated")
    return c


parser = ArgumentParser()
parser.add_argument('inputfile')
parser.add_argument('--json', help='print json', action='store_true')


def main():
    args = parser.parse_args()

    with EventIOFile(args.inputfile) as f:
        counter = count_versions(f)

    if args.json:
        object_info = [
            {'type': t, 'version': v, 'number_of_objects': c}
            for (t, v), c in sorted(list(counter.items()))
        ]
        print(json.dumps(object_info, indent=2))
    else:
        print(' Type | Version | #Objects | eventio-class')
        print('-' * 60)
        for (t, v, q), c in sorted(list(counter.items())):
            print('{: 5d} | {: 7d} | {: 8d} | {}'.format(
                t, v, c, q,
            ))
        print('-' * 60)


if __name__ == '__main__':
    main()



import gzip
try:
    import zstandard as zstd
    has_zstd = True
except ImportError:
    has_zstd = False

from .constants import (
    SYNC_MARKER_SIZE,
    SYNC_MARKER_LITTLE_ENDIAN,
    SYNC_MARKER_BIG_ENDIAN,
)


def is_gzip(path):
    '''Test if a file is gzipped by reading its first two bytes and compare
    to the gzip marker bytes.
    '''
    with open(path, 'rb') as f:
        marker_bytes = f.read(2)

    return marker_bytes[0] == 0x1f and marker_bytes[1] == 0x8b


def is_zstd(path):
    '''Test if a file is compressed using zstd using its magic marker bytes
    '''
    with open(path, 'rb') as f:
        marker_bytes = f.read(4)

    return marker_bytes == b'\x28\xb5\x2f\xfd'


def is_eventio(path):
    '''
    Test if a file is a valid eventio file by checking if the sync marker is there.
    '''
    if is_gzip(path):
        with gzip.open(path, 'rb') as f:
            marker_bytes = f.read(SYNC_MARKER_SIZE)
    elif is_zstd(path):
        if not has_zstd:
            raise IOError('You need the `zstandard` module to read zstd files')
        with open(path, 'rb') as f:
            cctx = zstd.ZstdDecompressor()
            with cctx.stream_reader(f) as stream:
                marker_bytes = stream.read(SYNC_MARKER_SIZE)
    else:
        with open(path, 'rb') as f:
            marker_bytes = f.read(SYNC_MARKER_SIZE)

    little = marker_bytes == SYNC_MARKER_LITTLE_ENDIAN
    big = marker_bytes == SYNC_MARKER_BIG_ENDIAN

    return little or big



import struct
import gzip
import logging

from .file_types import is_gzip, is_eventio, is_zstd
from .header import parse_header_bytes, get_bits_from_word
from . import constants
from .exceptions import WrongType

try:
    import zstandard as zstd
    has_zstd = True
except ImportError:
    has_zstd = False

log = logging.getLogger(__name__)

KNOWN_OBJECTS = {}


class EventIOFile:

    def __init__(self, path):
        log.info('Opening new file {}'.format(path))
        self.path = path

        if not is_eventio(path):
            raise ValueError('File {} is not an eventio file'.format(path))

        if is_gzip(path):
            log.info('Found gzipped file')
            self._filehandle = gzip.open(path, mode='rb')
        elif is_zstd(path):
            log.info('Found zstd compressed file')
            if not has_zstd:
                raise IOError(
                    'You need to install the `zstandard module'
                    'to read zstd compressed file`'
                )
            cctx = zstd.ZstdDecompressor()
            self._filehandle = cctx.stream_reader(open(path, 'rb'))
        else:
            log.info('Found uncompressed file')
            self._filehandle = open(path, mode='rb')

        self._next_header_pos = 0

    def __iter__(self):
        self._next_header_pos = 0
        return self

    def __next__(self):
        self._filehandle.seek(self._next_header_pos)
        header = read_next_header_toplevel(self)
        self._next_header_pos = self._filehandle.tell() + header.length
        return KNOWN_OBJECTS.get(header.type, EventIOObject)(
            header,
            filehandle=self._filehandle,
        )

    def seek(self, position, whence=0):
        return self._filehandle.seek(position, whence)

    def tell(self):
        return self._filehandle.tell()

    def read(self, size=-1):
        return self._filehandle.read(size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self._filehandle.close()


def check_size_or_raise(data, expected_length, zero_ok=True):
    if len(data) == 0:
        if zero_ok:
            raise StopIteration
        else:
            raise EOFError('File seems to be truncated')

    if len(data) < expected_length:
        raise EOFError('File seems to be truncated')


def read_next_header_toplevel(byte_stream):
    '''Read the next toplevel header object from the file
    Assumes position of `byte_stream` is at the beginning of a new header.

    Raises stop iteration if not enough data is available.
    '''
    sync = byte_stream.read(constants.SYNC_MARKER_SIZE)
    check_size_or_raise(sync, constants.SYNC_MARKER_SIZE)
    endianness = parse_sync_bytes(sync)

    return read_next_header_sublevel(byte_stream, endianness)


def read_next_header_sublevel(byte_stream, endianness, parent_address=0):
    '''Read the next sublevel header object from the file
    Assumes position of `byte_stream` is at the beginning of a new header.

    endianness: char
        '<' or '>' for little or big endiannes respectively.
        Is either read from next header or already known.

    Raises stop iteration if not enough data is available.
    '''
    if endianness == '>':
        raise NotImplementedError(
            'Big endian byte order is not supported by this reader'
        )

    header_bytes = byte_stream.read(constants.OBJECT_HEADER_SIZE)
    check_size_or_raise(
        header_bytes,
        constants.OBJECT_HEADER_SIZE,
        zero_ok=False,
    )

    header = parse_header_bytes(header_bytes)
    header.endianness = endianness

    if header.extended:
        extension_field = byte_stream.read(constants.EXTENSION_SIZE)
        check_size_or_raise(
            extension_field,
            constants.EXTENSION_SIZE,
            zero_ok=True
        )
        header.length += parse_extension_field(extension_field)

    header.address = parent_address + byte_stream.tell()

    return header


def parse_sync_bytes(sync):
    ''' returns the endianness as given by the sync byte '''

    if sync == constants.SYNC_MARKER_LITTLE_ENDIAN:
        log.debug('Found Little Endian byte order')
        return '<'

    if sync == constants.SYNC_MARKER_BIG_ENDIAN:
        log.debug('Found Big Endian byte order')
        return '>'

    raise ValueError(
        'Sync must be 0xD41F8A37 or 0x378A1FD4. Got: {}'.format(sync)
    )


class EventIOObject:
    '''
    Base Class for eventio objects.
    Can be subclassed to implement different types of
    EventIO objects and how their data payload is parsed into
    python objects.

    EventIO objects can basically play two roles:
        - a binary data blob
        - A list of other `EventIOObject`s

    If an `EventIOObject` is a pure list of other `EventIOObject`s,
    it can be iterated.
    Otherwise, parsing of the binary data has to be implented in a subclass
    or done "by hand" after reading the payload bytes.
    '''
    eventio_type = None

    def __init__(self, header, filehandle):
        if self.eventio_type is not None and header.type != self.eventio_type:
            raise WrongType(self.eventio_type, header.type)

        self._filehandle = filehandle
        self.header = header
        self.address = self.header.address
        self.length = self.header.length
        self.only_subobjects = self.header.only_subobjects
        self._next_header_pos = 0

    def read(self, size=-1):
        '''Read bytes from the payload of this object.

        Parameters
        ----------
        size: int
            read `size` bytes from the payload of this object
            If size == -1 (default), read all remaining bytes.
        '''
        pos = self.tell()

        # read all remaining bytes.
        if size == -1 or size > self.length - pos:
            size = self.length - pos

        data = self._filehandle.read(size)

        return data

    def __iter__(self):
        if not self.header.only_subobjects:
            raise ValueError(
                'Only EventIOObjects that contain just subobjects are iterable'
            )
        self._next_header_pos = 0
        return self

    def __next__(self):
        if not self.only_subobjects:
            raise ValueError(
                'Only EventIOObjects that contain just subobjects are iterable'
            )

        if self._next_header_pos >= self.header.length:
            raise StopIteration

        self.seek(self._next_header_pos)
        header = read_next_header_sublevel(
            self, self.header.endianness, parent_address=self.address
        )
        self._next_header_pos = self.tell() + header.length
        return KNOWN_OBJECTS.get(header.type, EventIOObject)(
            header, filehandle=self._filehandle
        )

    def seek(self, offset, whence=0):
        address = self.address
        if whence == 0:
            assert offset >= 0
            self._filehandle.seek(address + offset, whence)
        elif whence == 1:
            self._filehandle.seek(offset, whence)
        elif whence == 2:
            if offset > self.length:
                offset = self.length
            self._position = self._filehandle.seek(address + self.length - offset, 0)
        else:
            raise ValueError(
                'invalid whence ({}, should be 0, 1 or 2)'.format(whence)
            )
        return self.tell()

    def tell(self):
        return self._filehandle.tell() - self.address

    def __repr__(self):
        return '{}[{}](size={}, only_subobjects={}, address={})'.format(
            self.__class__.__name__,
            self.header.type,
            self.header.length,
            self.header.only_subobjects,
            self.header.address
        )


def parse_extension_field(extension_field):
    '''parse the so called "extension" field from `file`

    The length of an object can be so large, that it cannot be hold by the
    original `length` field which is 30bits long.
    In that case the most significant part of the length is stored in the
    so called "extension" field. The extension is 12bits long.

    So the total length of the object is:
        real_length = extension * 2^30 + original_lenth

    This function returns the extension *already multiplied with 2^30*
    so that the original length can simply be added to the result of this
    function in order to get the real length of the object.
    '''
    word, = struct.unpack('<I', extension_field)
    extension = get_bits_from_word(
        word, constants.EXTENSION_NUM_BITS, constants.EXTENSION_POS
    )

    # we push the length-extension so many bits to the left,
    # i.e. we multiply with such a high number, that we can simply
    # use the += operator  to combine the normal (small) length and this extension.
    extension <<= constants.LENGTH_NUM_BITS
    return extension
